// Copyright 2007 by Intuit Inc.
// All rights reserved
// Usage governed by the QuickBooks SDK Developer's License Agreement

using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Threading;
using System.Xml;
using Microsoft.Win32;
using Interop.QBFC10;
using COMEventHandler;
using Intuit_QBFC1.Subscription_Framework.Menu;
using Intuit_QBFC1.Subscription_Framework.Data;
using Intuit_QBFC1.Subscription_Framework.UI;
using Intuit_QBFC1.Session_Framework;

namespace Intuit_QBFC1
{
    static class Program
    {
        private static bool bSetupHandler = false;

        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main(string[] args)
        {
            try
            {
                VistaTools vistaTools = VistaTools.getInstance();
                bool doCOM = true;
                bool silentFlag = false;

                // Check for the -silent flag
                if (args.Length == 2)
                {
                    if ((args[0] == "-silent")  || (args[1] == "-silent"))
                        silentFlag = true;
                }

                for (int i = 0; i < args.Length; i++)
                {
                    switch (args[i])
                    {
                        case "-embedding": //COM sends this as the argument when starting an out of proc server
                            HandleQBEvent.embeddingOp(silentFlag);
                            break;

                        case "-regserver":
                        case "/regserver":
                            doCOM = false;
                            if (vistaTools.vistaModeCheck(true, args))
                                break;

                            HandleQBEvent.regserverOp(silentFlag);
                            break;

                        case "-unregserver":
                        case "/unregserver":
                            doCOM = false;
                            if (vistaTools.vistaModeCheck(true, args))
                                break;

                            HandleQBEvent.unregserverOp(silentFlag);
                            break;

                        case "-subscribe":
                            doCOM = false;
                            if (vistaTools.vistaModeCheck(false, args))
                                break;

                            subscribeMenuItems();
                            //TODO: START REMOVE - These will be removed in the next release!
                            (new DataSubscriptionManager(Defaults.SUBSCRIBER_ID, Defaults.CLASS_ID)).deleteDataSubscriptions();
                            (new UISubscriptionManager(Defaults.SUBSCRIBER_ID, Defaults.CLASS_ID)).deleteUISubscriptions();
                            //TODO: END REMOVE
                            subscribeDataEvents();
                            subscribeUiEvents();

                            // any errors would result in Exceptions being thrown.  Because Quickbooks
                            // must be restarted in order for new subscriptions to be available, set
                            // the doCOM flag to false.
                            if (!silentFlag)
                            {
                                if (vistaTools.isVista())
                                    MessageBox.Show("QuickBooks subscription was successful. QuickBooks must now be restarted!\n" +
                                                    "IMPORTANT NOTE:\n" +
                                                    "   QuickBooks must be restarted as an Administrator ONE TIME\n" +
                                                    "   for the menu subscriptions to be visible to the \"normal\" user!"
                                                   );
                                else
                                    MessageBox.Show("QuickBooks subscription was successful. Quickbooks must now be restarted!");
                            }
                            break;

                        case "-unsubscribe":
                            doCOM = false;
                            if (vistaTools.vistaModeCheck(false, args))
                                break;

                            setupQBEventObject();
                            (new MenuSubscriptionManager(Defaults.SUBSCRIBER_ID, Defaults.CLASS_ID)).deleteMenuSubscriptions();
                            (new DataSubscriptionManager(Defaults.SUBSCRIBER_ID, Defaults.CLASS_ID)).deleteDataSubscriptions();
                            (new UISubscriptionManager(Defaults.SUBSCRIBER_ID, Defaults.CLASS_ID)).deleteUISubscriptions();
                            if (!silentFlag)
                            {
                                if (vistaTools.isVista())
                                    MessageBox.Show("QuickBooks subscription removal was successful. QuickBooks must now be restarted!\n" +
                                                    "IMPORTANT NOTE:\n" +
                                                    "   QuickBooks must be restarted as an Administrator ONE TIME\n" +
                                                    "   for the menu subscriptions to be visible to the \"normal\" user!"
                                                   );
                                else
                                    MessageBox.Show("QuickBooks subscription removal was successful. Quickbooks must now be restarted!");
                            }
                            break;

                        case "-clearregistry":
                            doCOM = false;
                            if (vistaTools.vistaModeCheck(false, args))
                                break;

                            setupQBEventObject();
                            Registry.CurrentUser.DeleteSubKey(Defaults.REG_KEY, false);
                            if (!silentFlag)
                                MessageBox.Show("Registry data has been removed");
                            break;

                        default:
                            break;
                    }
                }
                if (doCOM && !vistaTools.vistaModeCheck(false, args))
                {
                    setupQBEventObject();
                    HandleQBEvent.StartCOM();
                    Application.EnableVisualStyles();
                    Application.SetCompatibleTextRenderingDefault(false);

                    // Start the application!
                    Application.Run(new Form1());
                    HandleQBEvent.StopCOM();
                    Logger.getInstance().logInfo("Application is closing normally");
                }
            }
            catch (Exception e)
            {
                Logger log = Logger.getInstance();
                log.logCritical("Program.Main", e.Message);
                log.logCritical("Program.Main", "Application Terminating");
                MessageBox.Show("An error requires the application to close:\n" +
                                "Logging File: " + Logger.getInstance().LogLevel.ToString() + "\n" +
                                "Log File: " + log.LogFile + "\n" +
                                "Exception Caught: \n   " + e.Message);
            }
        }

        private static void setupQBEventObject()
        {
            if (bSetupHandler == false)
            {
                SessionManager sessionMgr = SessionManager.getInstance();
                HandleQBEvent.setVersionInfo(sessionMgr.QBsdkMajorVersion, sessionMgr.QBsdkMinorVersion);
                bSetupHandler = true;
            }
        }

        public static void subscribeMenuItems()
        {
            $generated_menu_code$
        }
        
        public static void subscribeDataEvents()
        {
            $generated_data_code$
        }

        public static void subscribeUiEvents()
        {
            $generated_ui_code$
        }
    }
}