// Copyright 2007 by Intuit Inc.
// All rights reserved
// Usage governed by the QuickBooks SDK Developer's License Agreement

using System;
using System.Collections.Generic;
using System.Text;
using Interop.QBFC10;
using Intuit_QBFC1.Session_Framework;

namespace Intuit_QBFC1.Subscription_Framework.UI
{
    class UISubscriptionManager
    {
        private SessionManager sessionMgr = SessionManager.getInstance();   // Instance holder
        private ISubscriptionMsgSetRequest _msgSet = null;      // Holder for the subscription message set
        private IUIEventSubscriptionAdd _request = null;        // Holder for the request object
        private Guid _guid = Guid.Empty;                        // Holder for the subscriber GUID; initially empty
        private Guid _classId;                                  // Holder for the COM classID
        private string _progId;                                 // Holder for the COM progID
        private bool _bOnOpening;                               // Opening event
        private bool _bOnClosing;                               // Closing event
        private ENDeliveryPolicy _deliveryPolicy = ENDeliveryPolicy.dpDeliverOnlyIfRunning;

        // Don't allow the default constructor...
        private UISubscriptionManager() { }

        public UISubscriptionManager(Guid subscriberId, Guid classId)
            : this(subscriberId, null)
        {
            _classId = classId;
        }

        public UISubscriptionManager(Guid subscriberId, string progId)
        {
            _guid = subscriberId;
            _progId = progId;

            try
            {
                // go and determine if the backend is even capable of processing subscriptions...
                if (sessionMgr.QBsdkVersion < 3.0)
                {
                    Logger.getInstance().logError("UISubscriptionManager.UISubscriptionManager", "Backend is too old to support subscriptions!");
                    throw new UISubscriptionException("Opps! Your backend does not support subscriptions!");
                }

                // Next determine if the GUID is there...
                if (_guid == Guid.Empty)
                {
                    Logger.getInstance().logError("UISubscriptionManager.UISubscriptionManager", "A GUID for the application was not provided and is mandatory");
                    throw new UISubscriptionException("Opps! No GUID for the application was found!");
                }
            }
            catch (Exception e)
            {
                Logger.getInstance().logCritical("UISubscriptionManager.UISubscriptionManager", e.Message);
                throw e;
            }
        }

        /// <summary>
        /// Sets up the request objects...
        /// </summary>
        /// <returns>the ui event subscription object</returns>
        private IUIEventSubscriptionAdd getSubscriptionAddRq()
        {
            if (_msgSet == null)
            {
                _msgSet = sessionMgr.getSubscriptionMsgSetRequest();
                _request = null;
            }

            if (_request == null)
                _request = _msgSet.AppendUIEventSubscriptionAddRq();

            // Setup the request...
            _request.SubscriberID.SetValue(_guid.ToString("B"));
            _request.COMCallbackInfo.AppName.SetValue(sessionMgr.AppName);

            // Set the ProgID or ClassID, as appropriate
            if (_classId != Guid.Empty)
                _request.COMCallbackInfo.ORProgCLSID.CLSID.SetValue(_classId.ToString("B"));
            else if (_progId != null)
                _request.COMCallbackInfo.ORProgCLSID.ProgID.SetValue(_progId.ToString());
            else
            {
                Logger.getInstance().logError("UISubscriptionManager.getSubscriptionAddRq", "Either the COM classID or COM progID must be provided!");
                throw new UISubscriptionException("Error: Either the COM classID or COM progID must be provided!");
            }

            return _request;
        }

        /// <summary>
        /// Clean-up method that should always be called when done
        /// </summary>
        private void cleanUpSubscriptionAddRq()
        {
            _request = null;
            _msgSet = null;
        }

        /// <summary>
        /// Deletes any data event subscriptions for this application in QuickBooks
        /// </summary>
        public void deleteUISubscriptions()
        {
            try
            {
                // Create a message request set.
                _msgSet = sessionMgr.getSubscriptionMsgSetRequest();

                // get the query object, set the GUID and then execute the request
                ISubscriptionDel lQuery = _msgSet.AppendSubscriptionDelRq();
                lQuery.SubscriberID.SetValue(_guid.ToString("B"));
                lQuery.SubscriptionType.SetValue(ENSubscriptionType.stUI);
                sessionMgr.doSubscriptionRequests(ref _msgSet);
                _msgSet = null;
            }
            catch (Exception e)
            {
                Logger.getInstance().logCritical("UISubscriptionManager.deleteUISubscriptions", e.Message);
                throw e;
            }
        }

        /// <summary>
        /// Gets/Sets the event's delivery policy (see the OSR)
        /// </summary>
        public ENDeliveryPolicy DeliveryPolicy
        {
            get
            {
                return _deliveryPolicy;
            }
            set
            {
                _deliveryPolicy = value;
            }
        }

        /// <summary>
        /// Gets/Sets the policy on delivery of  events when QuickBooks closes (see the OSR)
        /// </summary>
        public bool ClosingEvent
        {
            get
            {
                return _bOnClosing;
            }
            set
            {
                _bOnClosing = value;
            }
        }

        /// <summary>
        /// Gets/Sets the policy on delivery of events when QuickBooks Opens (see the OSR)
        /// </summary>
        public bool OpeningEvent
        {
            get
            {
                return _bOnOpening;
            }
            set
            {
                _bOnOpening = value;
            }
        }

        /// <summary>
        /// Subscribes the previously setup data events
        /// </summary>
        public void subscribe()
        {
            try
            {
                getSubscriptionAddRq();
                if (_msgSet == null || _request == null)
                {
                    Logger.getInstance().logCritical("UISubscriptionManager.subscribe", "You must setup UI events before attempting to subscribe!");
                    throw new UISubscriptionException("Error: you must setup UI events before attempting to subscribe!");
                }

                if (!ClosingEvent && !OpeningEvent)
                {
                    Logger.getInstance().logCritical("UISubscriptionManager.subscribe", "Either an opening event or a closing event must be specified (see the OSR)!");
                    throw new UISubscriptionException("Error: either an opening event or a closing event must be specified (see the OSR)!");
                }

                // Add the request's global event information now
                _request.DeliveryPolicy.SetValue(_deliveryPolicy);

                if (OpeningEvent)
                    _request.CompanyFileEventSubscription.CompanyFileEventOperationList.Add(ENCompanyFileEventOperation.cfeoOpen);
                if (ClosingEvent)
                    _request.CompanyFileEventSubscription.CompanyFileEventOperationList.Add(ENCompanyFileEventOperation.cfeoClose);

                // And now subscribe!
                if (sessionMgr.doSubscriptionRequests(ref _msgSet))
                    throw new UISubscriptionException("Error encountered while attempting to perform UI Subscription!");

                // Always cleanup after yourself!
                cleanUpSubscriptionAddRq();
            }
            catch (Exception e)
            {
                Logger.getInstance().logCritical("UISubscriptionManager.subscribe", e.Message);
                throw e;
            }
        }
    }
}
